
/**
 * @file oh_fs.c
 * @version $Id$
 *
 * The source file that has the subset of POSIX filesystem functions.
 */

#include <windows.h>

#include "oh_port.h"
#include "oh_fs.h"
#include "oh_fsutl.h"
#include "oh_debug.h"

typedef struct oh_fs_win32dir_t
{
  HANDLE handle;
  WIN32_FIND_DATAA finddata;
  oh_fs_dirent_t dirent;
} oh_fs_win32dir_t;

/**
 * POSIX open function.
 */
oh_int oh_fs_open(const oh_char *path, oh_int flags)
{
  oh_char pathbuf[OH_FS_MAXPATHLEN];
  DWORD access = GENERIC_READ;
  DWORD disposition = 0;
  HANDLE rethandle;
  oh_int mflags;
  oh_int after_trunc = 0;
  oh_int before_creat = 0;

  OH_ASSERT(path != OH_NULL);

  if (flags & OH_FS_O_WRONLY)
    access = GENERIC_WRITE;
  if (flags & OH_FS_O_RDWR)
    access = GENERIC_WRITE | GENERIC_READ;

  mflags = flags & (OH_FS_O_CREAT | OH_FS_O_EXCL | OH_FS_O_TRUNC);

  switch (mflags)
  {
    case 0:
    case OH_FS_O_EXCL:
      disposition = OPEN_EXISTING;
      break;
    case OH_FS_O_TRUNC:
    case (OH_FS_O_EXCL | OH_FS_O_TRUNC):
      if (access == GENERIC_READ)
        disposition = OPEN_EXISTING;
      else
        disposition = TRUNCATE_EXISTING;
      break;
    case OH_FS_O_CREAT:
      disposition = OPEN_ALWAYS;
      break;
    case (OH_FS_O_CREAT | OH_FS_O_TRUNC):
      if (access != GENERIC_READ)
        after_trunc = 1;
      disposition = OPEN_ALWAYS;
      break;
    case (OH_FS_O_CREAT | OH_FS_O_EXCL):
      disposition = CREATE_NEW;
      break;
    case (OH_FS_O_CREAT | OH_FS_O_EXCL | OH_FS_O_TRUNC):
      before_creat = 1;
      if (access == GENERIC_READ)
        disposition = OPEN_EXISTING;
      else
        disposition = TRUNCATE_EXISTING;
      break;
    default:
      OH_ASSERT(OH_FALSE);
      break;
  }

  if (oh_c_strlen(path) > OH_FS_MAXPATHLEN - 1)
    return -1;
  oh_c_strcpy(pathbuf, path);

  oh_fu_unix2dos(pathbuf);

  if (before_creat)
  {
    rethandle =
      CreateFileA(pathbuf, access,
        FILE_SHARE_DELETE | FILE_SHARE_READ | FILE_SHARE_WRITE, OH_NULL,
        CREATE_NEW, FILE_ATTRIBUTE_NORMAL, OH_NULL);

    if (rethandle == INVALID_HANDLE_VALUE)
      return -1;

    CloseHandle(rethandle);
  }

  /* use FILE_FLAG_BACKUP_SEMANTICS to open directory, too. */
  rethandle =
    CreateFileA(pathbuf, access,
      FILE_SHARE_DELETE | FILE_SHARE_READ | FILE_SHARE_WRITE, OH_NULL,
      disposition, FILE_FLAG_BACKUP_SEMANTICS, OH_NULL);

  if (rethandle == INVALID_HANDLE_VALUE)
    return -1;

  if (after_trunc)
  {
    if (!SetEndOfFile(rethandle))
    {
      CloseHandle(rethandle);
      return -1;
    }
  }

  return (oh_int)rethandle;
}

/**
 * POSIX close function.
 */
oh_int oh_fs_close(oh_int fd)
{
  return (CloseHandle((HANDLE)fd) == 0);
}

/**
 * POSIX read function.
 */
oh_long oh_fs_read(oh_int fd, oh_void *buf, oh_ulong count)
{
  oh_ulong readcount;

  OH_ASSERT(buf != OH_NULL);

  if (ReadFile((HANDLE)fd, buf, count, &readcount, OH_NULL))
    return (oh_long)readcount;

  return -1;
}

/**
 * POSIX write function.
 */
oh_long oh_fs_write(oh_int fd, const oh_void *buf, oh_ulong count)
{
  oh_ulong writecount;

  OH_ASSERT(buf != OH_NULL);

  if (WriteFile((HANDLE)fd, buf, count, &writecount, OH_NULL))
    return (oh_long)writecount;

  return -1;
}

/**
 * POSIX lseek function.
 */
oh_long oh_fs_lseek(oh_int fd, oh_long offset, oh_int whence)
{
  switch (whence)
  {
    case OH_FS_SEEK_SET:
      whence = FILE_BEGIN;
      break;
    case OH_FS_SEEK_CUR:
      whence = FILE_CURRENT;
      break;
    case OH_FS_SEEK_END:
      whence = FILE_END;
      break;
    default:
      return -1;
      break;
  }

  return SetFilePointer((HANDLE)fd, offset, OH_NULL, whence);
}

/**
 * POSIX ftruncate function.
 */
oh_int oh_fs_ftruncate(oh_int fd, oh_long length)
{
  if (SetFilePointer((HANDLE)fd, length, OH_NULL, FILE_BEGIN) != 0xffffffff)
  {
    if (SetEndOfFile((HANDLE)fd))
    {
      return 0;
    }
  }
  return -1;
}

/**
 * POSIX fsync function.
 */
oh_int oh_fs_fsync(oh_int fd)
{
  if (FlushFileBuffers((HANDLE)fd))
    return 0;
  return -1;
}

/**
 * POSIX fstat function.
 */
oh_int oh_fs_fstat(oh_int fd, oh_fs_stat_t *buf)
{
  BY_HANDLE_FILE_INFORMATION hfi;

  OH_ASSERT(buf != OH_NULL);

  if (GetFileInformationByHandle((HANDLE)fd, &hfi))
  {
    /* ooops. hard to calculate the time */
    buf->st_mtime = 0;
    buf->st_size = hfi.nFileSizeLow;
    if (hfi.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
      buf->st_mode = OH_FS_S_IFDIR;
    else
      buf->st_mode = OH_FS_S_IFREG;

    return 0;
  }

  return -1;
}

/**
 * POSIX rename function.
 */
oh_int oh_fs_rename(const oh_char *oldpath, const oh_char *newpath)
{
  oh_char oldpathbuf[OH_FS_MAXPATHLEN];
  oh_char newpathbuf[OH_FS_MAXPATHLEN];

  OH_ASSERT(oldpath != OH_NULL);
  OH_ASSERT(newpath != OH_NULL);

  if (oh_c_strlen(oldpath) > OH_FS_MAXPATHLEN - 1)
    return -1;
  oh_c_strcpy(oldpathbuf, oldpath);

  if (oh_c_strlen(newpath) > OH_FS_MAXPATHLEN - 1)
    return -1;
  oh_c_strcpy(newpathbuf, newpath);

  oh_fu_unix2dos(oldpathbuf);
  oh_fu_unix2dos(newpathbuf);

  return (MoveFileA(oldpathbuf, newpathbuf) == 0);
}

/**
 * POSIX unlink function.
 */
oh_int oh_fs_unlink(const oh_char *path)
{
  oh_char pathbuf[OH_FS_MAXPATHLEN];

  OH_ASSERT(path != OH_NULL);

  if (oh_c_strlen(path) > OH_FS_MAXPATHLEN - 1)
    return -1;
  oh_c_strcpy(pathbuf, path);

  oh_fu_unix2dos(pathbuf);

  return (DeleteFileA(pathbuf) == 0);
}

/**
 * POSIX realpath function.
 */
oh_char *oh_fs_realpath(const oh_char *path, oh_char *respath, oh_ulong size)
{
  oh_char pathbuf[OH_FS_MAXPATHLEN];
  oh_char *dummy;
  DWORD ret;

  OH_ASSERT(path != OH_NULL);
  OH_ASSERT(respath != OH_NULL);

  if (oh_c_strlen(path) > OH_FS_MAXPATHLEN - 1)
    return OH_NULL;
  oh_c_strcpy(pathbuf, path);

  oh_fu_unix2dos(pathbuf);

  ret = GetFullPathNameA(pathbuf, size, respath, &dummy);
  if (ret > size || ret == 0)
  {
    respath[0] = '\0';
    return OH_NULL;
  }

  oh_fu_dos2unix(respath);

  return respath;
}

/**
 * POSIX getcwd function.
 */
oh_char *oh_fs_getcwd(oh_char *buf, oh_ulong size)
{
  DWORD ret;

  OH_ASSERT(buf != OH_NULL);

  ret = GetCurrentDirectoryA(size, buf);

  if (ret > size || ret == 0)
  {
    buf[0] = '\0';
    return OH_NULL;
  }

  oh_fu_dos2unix(buf);

  return buf;
}

/**
 * POSIX chdir function.
 */
oh_int oh_fs_chdir(const oh_char *path)
{
  oh_char pathbuf[OH_FS_MAXPATHLEN];

  OH_ASSERT(path != OH_NULL);

  if (oh_c_strlen(path) > OH_FS_MAXPATHLEN - 1)
    return -1;
  oh_c_strcpy(pathbuf, path);

  oh_fu_unix2dos(pathbuf);

  return (SetCurrentDirectoryA(pathbuf) == 0);
}

/**
 * POSIX mkdir function.
 */
oh_int oh_fs_mkdir(const oh_char *path)
{
  oh_char pathbuf[OH_FS_MAXPATHLEN];

  OH_ASSERT(path != OH_NULL);

  if (oh_c_strlen(path) > OH_FS_MAXPATHLEN - 1)
    return -1;
  oh_c_strcpy(pathbuf, path);

  oh_fu_unix2dos(pathbuf);

  return (CreateDirectoryA(pathbuf, OH_NULL) == 0);
}

/**
 * POSIX rmdir function.
 */
oh_int oh_fs_rmdir(const oh_char *path)
{
  oh_char pathbuf[OH_FS_MAXPATHLEN];

  OH_ASSERT(path != OH_NULL);

  if (oh_c_strlen(path) > OH_FS_MAXPATHLEN - 1)
    return -1;
  oh_c_strcpy(pathbuf, path);

  oh_fu_unix2dos(pathbuf);

  return (RemoveDirectoryA(pathbuf) == 0);
}

/**
 * POSIX opendir function.
 */
oh_fs_dir_t *oh_fs_opendir(const oh_char *path)
{
  oh_char pathbuf[OH_FS_MAXPATHLEN];
  oh_fs_win32dir_t *win32dir;

  OH_ASSERT(path != OH_NULL);

  if (oh_c_strlen(path) > OH_FS_MAXPATHLEN - 3) /* /, *, \0 */
    return OH_NULL;
  oh_c_strcpy(pathbuf, path);
  oh_c_strcat(pathbuf, "/*");

  oh_fu_unix2dos(pathbuf);

  win32dir = oh_c_malloc(sizeof(oh_fs_win32dir_t));
  if (win32dir == OH_NULL)
    return OH_NULL;

  win32dir->handle = FindFirstFileA(pathbuf, &win32dir->finddata);
  if (win32dir->handle == INVALID_HANDLE_VALUE)
  {
    oh_c_free(win32dir);
    return OH_NULL;
  }

  return (oh_fs_dir_t *)win32dir;
}

/**
 * POSIX readdir function.
 */
oh_fs_dirent_t *oh_fs_readdir(oh_fs_dir_t *dir)
{
  oh_fs_win32dir_t *win32dir;

  OH_ASSERT(dir != OH_NULL);

  win32dir = (oh_fs_win32dir_t *)dir;

  oh_c_strncpy(win32dir->dirent.d_name, win32dir->finddata.cFileName,
    OH_FS_MAXNAMELEN);
  win32dir->dirent.d_name[OH_FS_MAXNAMELEN - 1] = '\0';

  if (win32dir->dirent.d_name[0] != '\0')
  {
    if (FindNextFileA(win32dir->handle, &win32dir->finddata) == 0)
    {
      win32dir->finddata.cFileName[0] = '\0';
    }

    return &win32dir->dirent;
  }

  return OH_NULL;
}

/**
 * POSIX closedir function.
 */
oh_int oh_fs_closedir(oh_fs_dir_t *dir)
{
  HANDLE handle;

  OH_ASSERT(dir != OH_NULL);

  handle = ((oh_fs_win32dir_t *)dir)->handle;
  oh_c_free(dir);

  return (FindClose(handle) == 0);
}
