
/**
 * @file oh_debug.c
 * @version $Id$
 *
 * The source file that has the debug message control functions.
 *
 * All functions in the file are called only if OH_DEBUG is defined.
 */

#ifdef OH_DEBUG

#include "oh_port.h"
#include "oh_debug.h"
#include "oh_io.h"

/**
 * The maximum length of the debug message length.
 *
 * If the debug message is longer than it, then the message is truncated.
 */
#define _OH_DB_MAX_MSG_LEN 256

/**
 * The default debug message level.
 *
 * It is one of OL_ERR, OL_WRN, OL_MSG, or OL_DBG. It is usually used in the
 * _oh_db_nameLevels.
 */
#define _OH_DB_DEFAULT_LEVEL OL_DBG

typedef struct oh_db_name_level_t
{
  const oh_char *name;
  oh_db_level_t level;
} oh_db_name_level_t;

static oh_db_name_level_t _oh_db_nameLevels[] =
  {
    {"DEF",           _OH_DB_DEFAULT_LEVEL}, /*   0 */
    {"ASSERT",        _OH_DB_DEFAULT_LEVEL}, /*   1 */
    {"BOOT",          _OH_DB_DEFAULT_LEVEL}, /*   2 */
    {"PORT",          _OH_DB_DEFAULT_LEVEL}, /*   3 */
    {"NES",           _OH_DB_DEFAULT_LEVEL}, /*   4 */
    {"SYSTEM",        _OH_DB_DEFAULT_LEVEL}, /*   5 */
    {"EVENT",         _OH_DB_DEFAULT_LEVEL}, /*   6 */
    {"FILESYS",       _OH_DB_DEFAULT_LEVEL}, /*   7 */
    {"TESTKIT",       _OH_DB_DEFAULT_LEVEL}  /*   8 */
  };

static oh_uint _oh_db_numTypes = sizeof(_oh_db_nameLevels) /
  sizeof(oh_db_name_level_t);

/**
 * Prints the debug message.
 *
 * It has double parentheses because the macro usually doesn't support the
 * variable arguments. To remove oh_db_log() function call in the release mode,
 * the macro should recognize the variable arguments. To do that, make
 * oh_db_log() function have one argument to the macro with double parentheses.
 * The c preprocessor thinks that oh_db_log() has only one argument and the c
 * compiler thinks that oh_db_log() has the variable arguments.
 *
 * The higher debug message level implies the lower debug message level. In
 * other words, even if the current OD_FILESYS debug message level is OL_MSG,
 * OL_WRN or OL_ERR level debug message may be printed. The default debug
 * message level is defined in the _oh_db_nameLevels name level array and it
 * can be changed with the oh_db_setLevel() function.
 *
 * The variable argument, "format", is the same format with ANSI C printf
 * function argument.
 *
 * @param level The debug message level.
 * @param type The debug message type.
 * @param format The debug message.
 * @return none.
 */
oh_void oh_db_log((oh_db_level_t level, oh_db_type_t type,
  const oh_char *format, ...))
{
  va_list args;

  va_start(args, format);
  oh_db_vlog(level, type, format, args);
  va_end(args);
}

/**
 * Prints the debug message.
 *
 * It is same with oh_db_log() function except for it accepts va_list argument.
 * The "args" is the same format with ANSI C vprintf function argument.
 *
 * @param level The debug message level.
 * @param type The debug message type.
 * @param format The debug message.
 * @param args The additional argument of the "format".
 * @return none.
 */
oh_void oh_db_vlog(oh_db_level_t level, oh_db_type_t type,
  const oh_char *format, va_list args)
{
  static const oh_char *_oh_db_level[] =
    {
      "---", "ERR", "---", "WRN", "---", "---", "---", "MSG", "---", "---",
      "---", "---", "---", "---", "---", "DBG"
    };
  oh_char msg[_OH_DB_MAX_MSG_LEN];
  oh_char *msg_ptr;
  oh_char *msg_end_ptr;

  if (oh_db_checkLevel(type, level))
  {
    msg_ptr = msg;
    msg_end_ptr = msg_ptr + _OH_DB_MAX_MSG_LEN;

    oh_c_snprintf(msg_ptr, msg_end_ptr - msg_ptr, "%s|%s|",
      _oh_db_level[level], _oh_db_nameLevels[type].name);
    msg_ptr += oh_c_strlen(msg_ptr);
    oh_c_vsnprintf(msg_ptr, msg_end_ptr - msg_ptr, format, args);
    msg_ptr += oh_c_strlen(msg_ptr);
    oh_c_snprintf(msg_ptr, msg_end_ptr - msg_ptr, "\n");
    oh_io_putString(msg);
    oh_io_flush();
  }

  return;
}

/**
 * Gets the current debug message level of the specified debug message type.
 *
 * @param type The debug message type.
 * @return The current debug message level of the "type".
 */
oh_int oh_db_getLevel(oh_db_type_t type)
{
  if (((oh_uint)type) < _oh_db_numTypes)
    return _oh_db_nameLevels[type].level;
  else
  {
    oh_io_putString("ERR|DBG|illegal debug type in oh_db_get_level\n");
    return 0;
  }
}

/**
 * Sets the current debug message level of the debug message type.
 *
 * @param type The debug message type.
 * @param level The debug message level.
 * @return none.
 */
oh_void oh_db_setLevel(oh_db_type_t type, oh_db_level_t level)
{
  if (((oh_uint)type) < _oh_db_numTypes)
  {
    if (level == OL_ERR || level == OL_WRN || level == OL_MSG ||
      level == OL_DBG)
    {
      _oh_db_nameLevels[type].level = level;
    }
    else
    {
      oh_io_putString("ERR|DBG|illegal debug level in oh_db_set_level\n");
    }
  }
  else
  {
    oh_io_putString("ERR|DBG|illegal debug type in oh_db_set_level\n");
  }

  return;
}

/**
 * Checks if the message is printable with the specified "type", and "level".
 *
 * @param type The debug message type.
 * @param level The debug message level.
 * @return 0 if the message is not printable, not 0 otherwise.
 */
oh_int oh_db_checkLevel(oh_db_type_t type, oh_db_level_t level)
{
  if (((oh_uint)type) < _oh_db_numTypes)
  {
    if (level == OL_ERR || level == OL_WRN || level == OL_MSG ||
      level == OL_DBG)
    {
      return ((_oh_db_nameLevels[type].level & level) == level);
    }
    else
    {
      oh_io_putString("ERR|DBG|illegal debug level in oh_db_check_level\n");
    }
  }
  else
  {
    oh_io_putString("ERR|DBG|illegal debug type in oh_db_check_level\n");
  }

  return 0;
}

#endif /* OH_DEBUG */
